(*
    This file is part of the ParsingToolkit project -
    which provides combinator parsers for functional input streams.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    Authors: Fritz Henglein <henglein@it.edu>
             ANOQ of the Sun (alias Johnny Andersen)
               <anoq@HardcoreProcessing.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* Built upon Fritz Henglein's implementation of
   parser combinators as found in Larry Paulson's
   "ML for the Working Programmer"  *)

infix 6 $--
infix 6 --$
infix 5 --
infix 3 >>
(* infix 1 // *)
infix 0 ||

functor FParserCombinators(structure FIO : FUNCTIONAL_IO
                           type vec
                           type elem
                           val elemToString : elem -> string
                           val vecLength : vec -> int
                           val vecEqual : (vec * vec) -> bool
                           val vecToString : vec -> string
                           sharing type elem = FIO.elem
                           sharing type vec = FIO.vector)
          :> PARSER_COMBINATORS where type instream = FIO.instream
                                  and type vec = vec
                                  and type elem = elem =
  struct
    type instream = FIO.instream
    type vec = vec
    type elem = elem
    
    (* FIXME: Use FIO.StreamIO.PrimIO.compare instead... *)
    (* val csSize = Substring.size *)

    (* Exceptions for combinator errors *)
    exception SyntaxError of string * instream
    exception ValidityError of string

    type 'a parser = instream -> ('a * instream)

    fun (pf1 || pf2) stream =
          pf1 stream handle
            exn1 as (SyntaxError (_, stream1)) => 
              (pf2 stream handle
                 exn2 as (SyntaxError (_, stream2)) =>
                   raise exn1
                   (* FIXME: We could compare with FIO.StreamIO.PrimIO.compare if it was implemented *)
                   (* if csSize stream1 < csSize stream2 then
                     raise exn1
                   else
                     raise exn2 *) )

    fun (pf1 -- pf2) stream = 
          let
            val (res1, stream1) = pf1 stream
            val (res2, stream2) = pf2 stream1
          in
            ((res1, res2), stream2)
          end 

    fun (pf1 $-- pf2) stream = 
          let
            val (_, stream1) = pf1 stream
          in
            pf2 stream1
          end 

    fun (pf1 --$ pf2) stream = 
          let
            val (res1, stream1) = pf1 stream
            val (_, stream2) = pf2 stream1
          in
            (res1, stream2)
          end

    fun (pf >> f) stream =
          let
            val (res, stream') = pf stream
          in
            (f res, stream') handle
              ValidityError msg =>
                raise SyntaxError (msg, stream')
          end 

    fun $$ s stream =
          let
            val (v, stream2) = FIO.inputN (stream, vecLength s)
          in
            if vecEqual (v, s) then
              (s, stream2)
            else
              raise SyntaxError ((vecToString s) ^ " expected", stream)
          end

    fun empty stream = (nil, stream)

    (* Utility functions *)
    (* This implementation (and several others I've tried) runs
       out of memory on test.rib. I assume it is because of the
       stack of exceptions. *)
    (*
    fun repeat pf stream = 
          (pf -- repeat pf >> op:: || empty) stream *)

    (* Working implementation of repeat - takes 1.45 sec for parsing
       test.rib when using SML/NJ 110.0.6 on Linux on a 166Mhz Pentium.
       It does not use huge amounts of memory when compared to the amount
       of data being read. *)
    fun repeat pf stream =
          let
            fun oneIter stream =
                  let
                    val (res, stream') = pf stream
                  in
                    (SOME(res), stream')
                  end
                    handle SyntaxError _ =>
                      (NONE, stream)

            fun rep acc (NONE, stream) =
                  (rev acc, stream)
              | rep acc (SOME(res), stream) =
                  rep (res :: acc) (oneIter stream)
          in
            rep nil (oneIter stream)
          end

    fun repeatN n pf stream =
          if n > 0 then
            let
              fun oneIter stream =
                    let
                      val (res, stream') = pf stream
                    in
                      (SOME(res), stream')
                    end
                      handle SyntaxError _ =>
                        (NONE, stream)
  
              fun rep n acc (NONE, stream) =
                    raise SyntaxError ("Could not repeat parser the last " ^ (Int.toString n) ^ " times!", stream)
                | rep n acc (SOME(res), stream) =
                    if n > 0 then
                      rep (n - 1) (res :: acc) (oneIter stream)
                    else
                      (rev (res::acc), stream)
            in
              (* We subtract 1 from n because we call oneIter once already here *)
              rep (n - 1) nil (oneIter stream)
            end
          else
            (nil, stream)

    fun getIf pred stream =
          case FIO.input1 stream of
            SOME (res as (e, stream')) =>
              if pred e then
                res
              else
                raise SyntaxError ((elemToString e) ^ " unexpected", stream)
          | NONE =>
              raise SyntaxError ("Unexpected end of file", stream)

    fun repeatIf pred =
          repeat (getIf pred)

    fun repeatOneIf pred =
          getIf pred -- repeatIf pred >> op::
  end

structure TextIOParserCombinators =
  FParserCombinators(structure FIO = FuncTextIO
                     type vec = string
                     type elem = char
                     val elemToString = Char.toString
                     val vecLength = size
                     val vecEqual = (fn (s1, s2) => s1 = s2)
                     val vecToString = (fn s => s))

structure BinIOParserCombinators =
  FParserCombinators(structure FIO = FuncBinIO
                     type vec = Word8Vector.vector
                     type elem = Word8.word
                     (* FIXME: These anonymous functions are ugly...
                               implement them somewhere else *)
                     val elemToString = (fn e =>
                                           Char.toString
                                             (chr (Word8.toInt e)))
                     val vecLength = Word8Vector.length
                     val vecEqual = (fn (v1, v2) =>
                                       #2 (Word8Vector.foldl
                                             (fn (e, (index, value)) =>
                                                (index + 1,
                                                 value andalso ((Word8.compare (e, Word8Vector.sub (v2, index))) = EQUAL)))
                                             (0, true) v1))
                     val vecToString = (fn v =>
                                          String.concat
                                            (Word8Vector.foldr
                                               (fn (e, acc) => (elemToString e)::acc)
                                               nil v)))
