(*
    This file is part of the ParsingToolkit project -
    which provides combinator parsers for functional input streams.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    Authors: Fritz Henglein <henglein@it.edu>
             ANOQ of the Sun (alias Johnny Andersen)
               <anoq@HardcoreProcessing.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* Built upon Fritz Henglein's implementation of
   parser combinators as found in Larry Paulson's
   "ML for the Working Programmer"  *)

signature PARSER_COMBINATORS =
  sig
    type instream (* This is FuncTextIO.instream in TextIOParserCombinators
                         and FuncBinIO.instream in BinIOParserCombinators *)
    type vec (* This is string in TextIOParserCombinators
                    and Word8Vector.vector in BinIOParserCombinators *)
    type elem (* This is char in TextIOParserCombinators
                     and Word8.word in BinIOParserCombinators *)

    (* This is the type of a parser. A parser is a function
       taking a functional instream as argument. It returns
       a value that has been created by the parser during parsing,
       and a new functional instream with the stream position
       updated to where the parser stopped reading. *)
    type 'a parser = instream -> ('a * instream)

    (* SyntaxError is raised when a parser fails to parse. *)
    exception SyntaxError of string * instream

    (* Functions given to the >> combinator are expected to
       raise ValidityError on invalid arguments. *)
    exception ValidityError of string

    (* The combinators *)
    (* The purpose of this combinator is to try parsing with
       2 parser functions and return the result of the
       first function that succeeds. *)
    val || : ('a parser) * ('a parser) -> ('a parser)

    (* This combinator will execute 2 parsers in sequence and
       return a pair of the results of the parsers. *)
    val -- : ('a parser) * ('b parser) -> (('a * 'b) parser)

    (* This combinator executes 2 parsers in sequence and
       ignores the result of the first parser. *)
    val $-- : ('a parser) * ('b parser) -> ('b parser)

    (* This combinator executes 2 parsers in sequence and
       ignores the result of the second parser. *)
    val --$ : ('a parser) * ('b parser) -> ('a parser)

    (* Execute a parser and run the result through a function. *)
    val >> : ('a parser) * ('a -> 'b) -> ('b parser)

    (* This combinator is for reading an verifying an expected keyword. *)
    val $$ : vec -> (vec parser)

    (* Some handy built-in parsers. *)

    (* Doesn't parse anything, just returns nil. *)
    val empty : ('a list) parser

    (* Given a predicate, returns a parser that will read an
       element from the stream if the predicate is true. *)
    val getIf : (elem -> bool) -> (elem parser)

    (* Given a parser, returns a parser that will read a list
       of values with the given parser. Parses as many values as possible. *)
    val repeat : ('a parser) -> (('a list) parser)

    (* Given a predicate, returns a parser that will read a list of
       elements, until the predicate is false. *)
    val repeatIf : (elem -> bool) -> ((elem list) parser)

    (* Given a number n and a parser, returns a parser that
       parses a list of n values with the given parser. *)
    val repeatN : int -> ('a parser) -> ('a list parser)

    (* Same as repeatIf, except that this will read at least one
       value - or fail. *)
    val repeatOneIf : (elem -> bool) -> ((elem list) parser)
  end
